/*
****************************************************************************
****************************************************************************
**                                                                        **
**                XXXX     X    X         X XXXXXX X                      **
**               X    X   X X   X    X    X X      X   X                  **
**               X       X   X  X   X X   X X      X  X                   **
**               X   XX XXXXXXX X  X   X  X XXXX   X X                    **
**               X    X X     X X X     X X X      XX                     **
**               X    X X     X XX       XX X      X                      **
**                XXXX  X     X X         X XXXXXX XXXXXX                 **
**                                                                        **
**                           BIAYSTOK 2018                               **
**                                                                        **
**                         Home Control System                            **
**                                                                        **
**                             HC_MCTRL1                                  **
**                      Wersja z dnia : 2018.09.01                        **
**                     Napisal : Andrzej Pawluczuk                        **
**                                                                        **
****************************************************************************
****************************************************************************
*/

#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <avr/wdt.h>
#include "mctrl1eeprom.h"
#include "timeeventmodule.h"


#define false                          0
#define true                           1

#define PowerOnOffLed                  2 // 1
#define StatusLed                      3 // 0
#define RxTxLed                        1 // 2
#define KeyLed                         0 // 3

#define KeybPort                       PORTD
#define KeybPortDirection              DDRD
#define InpKeybReg                     PIND

#define LKey                           5              // PORTD 5

#define TimerCounterDelayLimit         3000
#define StandardLEDPulseDelay          832
#define TimerPrescaller                3600
#define AutoOnMotorTimeInterval        7000000
#define AutoOffMotorTimeInterval       100

#define LEDPort                        PORTC
#define LEDPortDirection               DDRC

#define PowerPort                      PORTB
#define PowerPortDirection             DDRB

#define PwrSwitchPin                   0

#define RxDPin                         0              // PORTD . 0
#define TxDPin                         1              // PORTD . 1
#define RxTxDirPin                     2              // PORTD . 2

#define null                           0
#define cr                             0x0D
#define Space                          0x20
#define PartSeparator                  ':'
#define ParamSeparator                 '.'

#define nop() __asm__ __volatile__ ("nop")

// USART BAUD Rate Generator fosc=7.3728 MHz
#define Speed9600bps                   47
#define Speed19200bps                  23
#define Speed38400bps                  11
#define Speed115200bps                 3

#define IdleState                      0
#define BusyState                      1

#define TXDBuffSize                    128
#define RXDBuffSize                    32

typedef struct { uint8_t CycFr ;
                 uint8_t CycTo ;
                 uint8_t State ;
                 uint8_t Buffer [ TXDBuffSize ] ;
               } TXDRecT ;
typedef struct { uint8_t CycFr ;
                 uint8_t CycTo ;
                 uint8_t Buffer [ RXDBuffSize ] ;
               } RXDRecT ;

#define CyclBuffSize                   256

typedef struct { uint8_t CycFr ;
                 uint8_t CycTo ;
                 uint8_t Buffer [ CyclBuffSize ] ;
               } CyclicRecT ;

typedef void ( * OperationProcT ) ( void ) ;

typedef struct {
                 volatile TXDRecT          TXDRec ;
                 volatile RXDRecT          RXDRec ;
                 OperationProcT            SerialService ;
                 CommandBufferType         CommandText ;
                 ParamBufferType           ParamText ;
                 DeviceIDType              DestinationDeviceName ;
                 DeviceIDType              SourceDeviceName ;
                 uint8_t                   CommandTextIndex ;
                 uint8_t                   ParamTextIndex ;
                 uint8_t                   DestDeviceNameIndex ;
                 uint8_t                   SrcDeviceNameIndex ;
                 uint16_t                  SerialLEDDelay ;
               } ServiceType ;

typedef ServiceType * PtrServiceType ;

#define TempEchoDevicesArraySize       4

static ServiceType SerialService ;
static uint8_t ConnectContext ;
static uint8_t TimeEvent ;
static uint16_t TimerCounter ;
static uint8_t MotorOnFlag ;
static DeviceIDType TempEchoDevices [ TempEchoDevicesArraySize ] ;
static CyclicRecT CyclicBuffer ;
static uint32_t ServiceDelayTime ;
static uint16_t LoTempBorder ;
static uint16_t HiTempBorder ;
static uint8_t StandardControlEnable ;
static PtrTimeEventElementType MotorOnHandler ;
static PtrTimeEventElementType MotorOffHandler ;

static const uint16_t CRCTable [ 16 ] = { 0x0000 ,  /* 00 */
                                          0xCC01 ,  /* 01 */
                                          0xD801 ,  /* 02 */
                                          0x1400 ,  /* 03 */
                                          0xF001 ,  /* 04 */
                                          0x3C00 ,  /* 05 */
                                          0x2800 ,  /* 06 */
                                          0xE401 ,  /* 07 */
                                          0xA001 ,  /* 08 */
                                          0x6C00 ,  /* 09 */
                                          0x7800 ,  /* 10 */
                                          0xB401 ,  /* 11 */
                                          0x5000 ,  /* 12 */
                                          0x9C01 ,  /* 13 */
                                          0x8801 ,  /* 14 */
                                          0x4400 } ;/* 15 */

static uint8_t DeviceName [ ] PROGMEM               = "HC-MCTRL - sterowanie pompa C.O.\r" ;
static uint8_t HardwareSoftwareVersion [ ] PROGMEM  = "HVER*1_0*IX-2011.SVER*1_0*IX-2018\r" ;
static uint8_t GeneralPoolDeviceName [ ] PROGMEM    = "GENERAL" ;
static uint8_t NoDataText [ ] PROGMEM                = "NODATA\r" ;
static uint8_t OK [ ] PROGMEM                       = "<OK>\r" ;
static uint8_t Fail [ ] PROGMEM                     = "<FAIL>\r" ;
static uint8_t EmptyFlagTxt [ ] PROGMEM              = "<EMPTY>\r" ;

static uint8_t ChannelIsOnState [ ] PROGMEM          = "CHISON\r" ;
static uint8_t ChannelIsOffState [ ] PROGMEM         = "CHISOF\r" ;

static uint8_t ReplyCommand [ ] PROGMEM              = "REPLY" ;
static uint8_t ConnectCommand [ ] PROGMEM            = "CONNECT" ;
static uint8_t DisConnectCommand [ ] PROGMEM         = "DISCONNECT" ;
static uint8_t SetDeviceNameCommand [ ] PROGMEM      = "SETMID" ;
static uint8_t HelloCommand [ ] PROGMEM              = "HELLO" ;
static uint8_t GetVersionCommand [ ] PROGMEM         = "GETVER" ;
static uint8_t GetIDCommand [ ] PROGMEM              = "GETID" ;
static uint8_t ReinitCommand [ ] PROGMEM             = "REINIT" ;
static uint8_t ResetEECommand [ ] PROGMEM            = "RESEE" ;
static uint8_t SaveEECommand [ ] PROGMEM             = "SAVEEE" ;
static uint8_t GetChnStatusCommand [ ] PROGMEM       = "GETCHNST" ;
static uint8_t AddSuperiorChnCommand [ ] PROGMEM     = "ADDSUP" ;
static uint8_t DelSuperiorChnCommand [ ] PROGMEM     = "DELSUP" ;
static uint8_t ListSuperiorChnCommand [ ] PROGMEM    = "LISTSUP" ;
static uint8_t ChStateServiceOnCommand [ ] PROGMEM   = "CSSERVON" ;
static uint8_t ChStateServiceOffCommand [ ] PROGMEM  = "CSSERVOFF" ;
static uint8_t ChStateServiceListCommand [ ] PROGMEM = "CSSERVLIST" ;
static uint8_t NewPasswordCommand [ ] PROGMEM        = "NEWPW" ;
static uint8_t SetTempThresholdCommand [ ] PROGMEM   = "STTHR" ;
static uint8_t GetTempThresholdCommand [ ] PROGMEM   = "GTTHR" ;
static uint8_t SetTempHysteresisCommand [ ] PROGMEM  = "STHIST" ;
static uint8_t GetTempHysteresisCommand [ ] PROGMEM  = "GTHIST" ;
static uint8_t AddParamCommand [ ] PROGMEM           = "ADDP" ;
static uint8_t DelParamCommand [ ] PROGMEM           = "DELP" ;
static uint8_t ListParamCommand [ ] PROGMEM          = "LISTP" ;
static uint8_t NewTempMeansCommand [ ] PROGMEM       = "RPV" ;


typedef void ( * CommandExecuteServiceProcT ) ( void ) ;
typedef struct {
                 CommandExecuteServiceProcT  CommandService ;
                 uint8_t *                   CommandName ;
               } CommandNameAssociationRecT ;

static void ReplyCommandExecute ( void ) ;
static void ConnectCommandExecute ( void ) ;
static void DisConnectCommandExecute ( void ) ;
static void SetDeviceNameCommandExecute ( void ) ;
static void HelloCommandExecute ( void ) ;
static void GetVersionCommandExecute ( void ) ;
static void GetIDCommandExecute ( void ) ;
static void ReinitCommandExecute ( void ) ;
static void ResetEECommandExecute ( void ) ;
static void SaveEECommandExecute ( void ) ;
static void GetChnStatusCommandExecute ( void ) ;
static void AddSuperiorChnCommandExecute ( void ) ;
static void DelSuperiorChnCommandExecute ( void ) ;
static void ChStateServiceOnCommandExecute ( void ) ;
static void ChStateServiceOffCommandExecute ( void ) ;
static void ChStateServiceListCommandExecute ( void ) ;
static void ListSuperiorChnCommandExecute ( void ) ;
static void NewPasswordCommandExecute ( void ) ;
static void SetTempThresholdCommandExecute ( void ) ;
static void GetTempThresholdCommandExecute ( void ) ;
static void SetTempHysteresisCommandExecute ( void ) ;
static void GetTempHysteresisCommandExecute ( void ) ;
static void AddParamCommandExecute ( void ) ;
static void DelParamCommandExecute ( void ) ;
static void ListParamCommandExecute ( void ) ;
static void NewTempMeansCommandExecute ( void ) ;


#define CommandNameAssociationTabeSize               26

static CommandNameAssociationRecT CommandNameAssociationTabe [ CommandNameAssociationTabeSize ] PROGMEM =
{
/* 00 */ { ReplyCommandExecute                 , ReplyCommand } ,
/* 01 */ { ConnectCommandExecute               , ConnectCommand } ,
/* 02 */ { DisConnectCommandExecute            , DisConnectCommand } ,
/* 03 */ { SetDeviceNameCommandExecute         , SetDeviceNameCommand } ,
/* 04 */ { HelloCommandExecute                 , HelloCommand } ,
/* 05 */ { GetVersionCommandExecute            , GetVersionCommand } ,
/* 06 */ { GetIDCommandExecute                 , GetIDCommand } ,
/* 07 */ { ReinitCommandExecute                , ReinitCommand } ,
/* 08 */ { ResetEECommandExecute               , ResetEECommand } ,
/* 09 */ { SaveEECommandExecute                , SaveEECommand } ,
/* 10 */ { GetChnStatusCommandExecute          , GetChnStatusCommand } ,
/* 11 */ { AddSuperiorChnCommandExecute        , AddSuperiorChnCommand } ,
/* 12 */ { DelSuperiorChnCommandExecute        , DelSuperiorChnCommand } ,
/* 13 */ { ListSuperiorChnCommandExecute       , ListSuperiorChnCommand } ,
/* 14 */ { ChStateServiceOnCommandExecute      , ChStateServiceOnCommand } ,
/* 15 */ { ChStateServiceOffCommandExecute     , ChStateServiceOffCommand } ,
/* 16 */ { ChStateServiceListCommandExecute    , ChStateServiceListCommand } ,
/* 17 */ { NewPasswordCommandExecute           , NewPasswordCommand } ,
/* 18 */ { SetTempThresholdCommandExecute      , SetTempThresholdCommand } ,
/* 19 */ { GetTempThresholdCommandExecute      , GetTempThresholdCommand } ,
/* 20 */ { SetTempHysteresisCommandExecute     , SetTempHysteresisCommand } ,
/* 21 */ { GetTempHysteresisCommandExecute     , GetTempHysteresisCommand } ,
/* 22 */ { AddParamCommandExecute              , AddParamCommand } ,
/* 23 */ { DelParamCommandExecute              , DelParamCommand } ,
/* 24 */ { ListParamCommandExecute             , ListParamCommand } ,
/* 25 */ { NewTempMeansCommandExecute          , NewTempMeansCommand } ,
} ;

static void USARTReceiveMode ( void )
{
  /*----------------------------------------------------------------*/
  DDRD |= ( 1 << RxTxDirPin ) ;
} /* USARTReceiveMode */


static void USARTTransmitMode ( void )
{
  /*----------------------------------------------------------------*/
  DDRD &= ~ ( 1 << RxTxDirPin ) ;
} /* USARTTransmitMode */


SIGNAL ( TIMER0_OVF_vect )
{
  /*----------------------------------------------------------------*/
  TimerCounter ++ ;
  if ( TimerCounter > TimerCounterDelayLimit )
  {
    SysTimerTickEvent ( ) ;
    TimerCounter = 0 ;
    TimeEvent = true ;
  } /* if */ ;
  if ( SerialService . SerialLEDDelay )
  {
    SerialService . SerialLEDDelay -- ;
    if ( ! SerialService . SerialLEDDelay )
    {
      LEDPort |= ( 1 << RxTxLed ) ;
    } /* if */ ;
  } /* if */ ;
} /* TIMER0_OVF_vect */


static void StartRxTxLED ( void )
{
  /*----------------------------------------------------------------*/
  LEDPort &= ~ ( 1 << RxTxLed ) ;
  SerialService . SerialLEDDelay = StandardLEDPulseDelay ;
} /* StartRxTxLED */


SIGNAL ( USART_RX_vect )
{
  uint8_t RecCh ;
  /*----------------------------------------------------------------*/
  RecCh = UDR0 ;
  SerialService . RXDRec . Buffer [ SerialService . RXDRec . CycTo ] = RecCh ;
  SerialService . RXDRec . CycTo ++ ;
  if ( SerialService . RXDRec . CycTo >= RXDBuffSize )
    SerialService . RXDRec . CycTo = 0 ;
  StartRxTxLED ( ) ;
} /* USART_RX_vect */


SIGNAL ( USART_TX_vect )
{
  /*----------------------------------------------------------------*/
  if ( SerialService . TXDRec . CycFr == SerialService . TXDRec . CycTo )
  {
    SerialService . TXDRec . State = IdleState ;
    USARTReceiveMode ( ) ;
  } /* if ... */
  else
  {
    UDR0 = SerialService . TXDRec . Buffer [ SerialService . TXDRec . CycFr ] ;
    SerialService . TXDRec . CycFr ++ ;
    if ( SerialService . TXDRec . CycFr >= TXDBuffSize )
      SerialService . TXDRec . CycFr = 0 ;
  } /* if ... else */ ;
} /* USART_TX_vect */


static void SendSerial ( uint8_t SerialCh )
{
  uint8_t NextInx ;
  /*----------------------------------------------------------------*/
  USARTTransmitMode ( ) ;
  StartRxTxLED ( ) ;
  cli ( ) ;
  if ( SerialService . TXDRec . State == IdleState )
  {
    SerialService . TXDRec . State = BusyState ;
    UDR0 = SerialCh ;
    sei ( ) ;
  } /* if ... */
  else
  {
    for ( ; ; )
    {
      NextInx = SerialService . TXDRec . CycTo + 1 ;
      if ( NextInx >= TXDBuffSize )
        NextInx = 0 ;
      if ( NextInx != SerialService . TXDRec . CycFr )
        break ;
      sei ( ) ;
      nop ( ) ;
      nop ( ) ;
      wdt_reset ( ) ;
      cli ( ) ;
    } /* for */ ;
    SerialService . TXDRec . Buffer [ SerialService . TXDRec . CycTo ] = SerialCh ;
    SerialService . TXDRec . CycTo ++ ;
    if ( SerialService . TXDRec . CycTo >= TXDBuffSize )
      SerialService . TXDRec . CycTo = 0 ;
    sei ( ) ;
  } /* if ... else */ ;
} /* SendSerial */


static void SendSerialTxt ( uint8_t * String )
{
  uint8_t Data ;
  /*----------------------------------------------------------------*/
  for ( ; ; )
  {
    Data = * String ++ ;
    if ( Data )
      SendSerial ( Data ) ;
    else
      break ;
  } /* for */ ;
} /* SendSerialTxt */


static void SendSerialFlashTxt ( uint16_t String )
{
  uint8_t Data ;
  /*----------------------------------------------------------------*/
  for ( ; ; )
  {
    Data = pgm_read_byte_near ( String ) ;
    if ( Data )
      SendSerial ( Data ) ;
    else
      break ;
    String ++ ;
  } /* for */ ;
} /* SendSerialFlashTxt */


static uint8_t SerialDataPresent ( void )
{
  /*----------------------------------------------------------------*/
  if ( SerialService . RXDRec . CycFr == SerialService . RXDRec . CycTo )
    return ( false ) ;
  else
    return ( true ) ;
} /* SerialDataPresent */


static uint8_t GetSerial ( void )
{
  uint8_t Ch ;
  /*----------------------------------------------------------------*/
  Ch = SerialService . RXDRec . Buffer [ SerialService . RXDRec . CycFr ] ;
  SerialService . RXDRec . CycFr ++ ;
  if ( SerialService . RXDRec . CycFr >= RXDBuffSize )
    SerialService . RXDRec . CycFr = 0 ;
  return ( Ch ) ;
} /* GetSerial */


static void AddToCyclic ( uint8_t Ch )
{
  /*----------------------------------------------------------------*/
  CyclicBuffer . Buffer [ CyclicBuffer . CycTo ] = Ch ;
  CyclicBuffer . CycTo ++ ;
//  if ( CyclicBuffer . CycTo >= CyclBuffSize )
//    CyclicBuffer. CycTo = 0 ;
} /* AddToCyclic */


static void AddToCyclicTxt ( uint8_t * String )
{
  uint8_t Data ;
  /*----------------------------------------------------------------*/
  for ( ; ; )
  {
    Data = * String ++ ;
    if ( Data )
      AddToCyclic ( Data ) ;
    else
      break ;
  } /* for */ ;
} /* AddToCyclicTxt */


static void AddToCyclicFlashTxt ( uint16_t String )
{
  uint8_t Data ;
  /*----------------------------------------------------------------*/
  for ( ; ; )
  {
    Data = pgm_read_byte_near ( String ) ;
    if ( Data )
      AddToCyclic ( Data ) ;
    else
      break ;
    String ++ ;
  } /* for */ ;
} /* AddToCyclicFlashTxt */


static uint8_t CyclicDataPresent ( void )
{
  /*----------------------------------------------------------------*/
  if ( CyclicBuffer . CycFr == CyclicBuffer . CycTo )
    return ( false ) ;
  else
    return ( true ) ;
} /* CyclicDataPresent */


static uint8_t GetFromCyclic ( void )
{
  uint8_t Ch ;
  /*----------------------------------------------------------------*/
  Ch = CyclicBuffer . Buffer [ CyclicBuffer . CycFr ] ;
  CyclicBuffer . CycFr ++ ;
//  if ( CyclicBuffer . CycFr >= RXDBuffSize )
//    CyclicBuffer . CycFr = 0 ;
  return ( Ch ) ;
} /* GetFromCyclic */


static void FlushCyclic ( void )
{
  uint8_t Ch ;
  /*----------------------------------------------------------------*/
  while ( CyclicDataPresent ( ) )
  {
    Ch = GetFromCyclic ( );
    SendSerial ( Ch ) ;
  } /* while */ ;
} /* FlushCyclic */


static void USARTHardwareInit ( uint16_t Speed )
{
  UCSR0A = 0 ;
  UCSR0B = ( 1 << RXCIE0 ) | ( 1 << TXCIE0 ) | ( 1 << RXEN0 ) | ( 1 << TXEN0 ) ;
  UCSR0C = ( 1 << UCSZ01 ) | ( 1 << UCSZ00 ) ;
  UBRR0 = Speed ;
} /* USARTHardwareInit */


static void HardwareInit ( uint16_t Speed )
{
  /*----------------------------------------------------------------*/
  TCCR0A = 0 ;
  TCCR0B = ( 1 << CS01 ) ; // CS02=0 CS01=1 CS00=0 : F clk div 8
  TCNT0 = 0 ;
  OCR0A = 0 ;
  OCR0B = 0 ;
  TIMSK0 = ( 1 << TOIE0 ) ;
  DDRD &= ~ ( 1 << RxDPin ) ;
  DDRD |= ( 1 << TxDPin ) | ( 1 << RxTxDirPin ) ;
  USARTHardwareInit ( Speed ) ;
} /* HardwareInit */


static void EnvirInit ( void )
{
  uint8_t Bits ;
  /*----------------------------------------------------------------*/
  Bits = ( 1 << PowerOnOffLed ) | ( 1 << StatusLed ) |
         ( 1 << RxTxLed ) | ( 1 << KeyLed ) ;
  LEDPortDirection |= Bits ;
  LEDPort |= Bits ;
  Bits = ( 1 << PwrSwitchPin ) ;
  PowerPortDirection |= Bits ;
  PowerPort |= Bits ;
} /* EnvirInit */


static void ClearBlock ( void *   Destination ,
                         uint16_t Size )
{
  uint16_t Loop ;
  uint8_t * Dst ;
  /*----------------------------------------------------------------*/
  Dst = ( uint8_t * ) Destination ;
  for ( Loop = 0 ; Loop < Size ; Loop ++ )
    * Dst ++ = null ;
} /* ClearBlock */


static void StrCopy ( uint8_t * Destination ,
                      uint8_t * Source )
{
  uint8_t Ch ;
  /*----------------------------------------------------------------*/
  Ch = * Source ;
  * Destination = Ch ;
  while ( Ch )
  {
    Ch = * Source ++ ;
    * Destination ++ = Ch ;
  } /* while */ ;
} /* StrCopy */


static void Assign ( uint8_t * SubDeviceID )
{
  uint8_t Loop ;
  uint8_t Ch ;
  /*----------------------------------------------------------------*/
  ClearBlock ( SubDeviceID , sizeof ( DeviceIDType ) ) ;
  for ( Loop = 0 ; Loop < DeviceIDSize ; Loop ++ )
  {
    Ch = SerialService . ParamText [ Loop ] ;
    if ( ! Ch )
      break ;
    * SubDeviceID ++ = Ch ;
  } /* for */ ;
} /* Assign */


static uint8_t CompareFlash ( uint8_t * Text ,
                              uint16_t  Pattern )
{
  uint8_t Ch1 ;
  uint8_t Ch2 ;
  /*----------------------------------------------------------------*/
  for ( ; ; )
  {
    Ch1 = * Text ++ ;
    Ch2 = pgm_read_byte_near ( Pattern ++ ) ;
    if ( Ch1 != Ch2 )
    {
      return ( false ) ;
    } /* if */ ;
    if ( Ch1 == 0 )
    {
      return ( true ) ;
    } /* if */ ;
  } /* for */ ;
} /* CompareFlash */


static uint8_t CompareDeviceID ( uint8_t * DeviceName ,
                                 uint8_t * Pattern )
{
  uint8_t Ch1 ;
  uint8_t Ch2 ;
  /*----------------------------------------------------------------*/
  for ( ; ; )
  {
    Ch1 = * DeviceName ++ ;
    Ch2 = * Pattern ++ ;
    if ( Ch1 != Ch2 )
    {
      return ( false ) ;
    } /* if */ ;
    if ( ( Ch1 == null ) || ( Ch2 == null ) )
      return ( true ) ;
  } /* for */ ;
} /* CompareDeviceID */


static void CodePassword ( void )
{
  uint8_t Loop ;
  uint8_t Element ;
  uint8_t BitCorrection ;
  /*----------------------------------------------------------------*/
  for ( Loop = 0 ; Loop <= PasswordSize ; Loop ++ )
  {
    Element = EEBlock . Password [ Loop ] ^ 0xFF ;
    BitCorrection = Element & 0x01 ;
    Element = Element >> 1 ;
    if ( BitCorrection )
      Element |= 0x80 ;
    EEBlock . Password [ Loop ] = Element ;
  } /* for */ ;
} /* CodePassword */


static void DecodePassword ( void )
{
  uint8_t Loop ;
  uint16_t Element ;
  /*----------------------------------------------------------------*/
  for ( Loop = 0 ; Loop <= PasswordSize ; Loop ++ )
  {
    Element = ( uint16_t ) EEBlock . Password [ Loop ] ;
    Element = Element << 1 ;
    if ( Element >= 0x100 )
      Element ++ ;
    Element &= 0xFF ;
    Element ^= 0xFF ;
    EEBlock . Password [ Loop ] = ( uint8_t ) Element ;
  } /* for */ ;
} /* DecodePassword */


static void ReadEEBlock ( void )
{
  /*----------------------------------------------------------------*/
  eeprom_read_block ( & EEBlock , ( void * ) EEBlockAddressOffset ,
                      sizeof ( EEPROMBlockDataRecT ) ) ;
} /* ReadEEBlock */


static uint16_t ComputeCRC ( uint16_t  Length ,
                             uint8_t * Buffer )
{
  uint16_t Loop ;
  uint16_t ResultCRC ;
  uint16_t WorkCRC ;
  /*----------------------------------------------------------------*/
  ResultCRC = 0xFFFF ;
  for ( Loop = 0 ; Loop < Length ; Loop ++ )
  {
    WorkCRC = CRCTable [ ResultCRC & 0x0F ] ;
    ResultCRC = ( ResultCRC >> 4 ) & 0x0FFF ;
    ResultCRC ^= WorkCRC ^ CRCTable [ Buffer [ Loop ] & 0x0F ] ;
    WorkCRC = CRCTable [ ResultCRC & 0x0F ] ;
    ResultCRC = ( ResultCRC >> 4 ) & 0x0FFF ;
    ResultCRC ^= WorkCRC ^ CRCTable [ ( Buffer [ Loop ] >> 4 ) & 0x0F ] ;
  } /* for */ ;
  return ( ResultCRC ) ;
} /* ComputeCRC */


static void WriteEEBlock ( void )
{
  uint16_t LocalCRCValue ;
  /*----------------------------------------------------------------*/
  CodePassword ( ) ;
  wdt_disable ( ) ;
  EEBlock . CRCValue = 0 ;
  LocalCRCValue = ComputeCRC ( sizeof ( EEPROMBlockDataRecT ) -
                          sizeof ( uint16_t ) ,
                          ( uint8_t * ) ( & EEBlock ) ) ;
  EEBlock . CRCValue = LocalCRCValue ;
  eeprom_write_block ( & EEBlock , ( void * ) EEBlockAddressOffset ,
                       sizeof ( EEPROMBlockDataRecT ) ) ;
  DecodePassword ( ) ;
  wdt_enable ( WDTO_1S ) ;
} /* WriteEEBlock */


static void WordToStr ( uint8_t * Number ,
                        uint8_t   ConvWidth ,
                        uint8_t   EmptyCh ,
                        uint16_t  Value )
{
  uint8_t Loop ;
  uint8_t LocalBuffer [ 8 ] ;
  uint8_t LastDigit ;
  /*----------------------------------------------------------------*/
  if ( ConvWidth > 7 )
  {
    * Number ++ = '!' ;
    * Number = ( uint8_t ) 0 ;
    return ;
  } /* if */ ;
  for ( Loop = 0 ; Loop < ConvWidth ; Loop ++ )
    LocalBuffer [ Loop ] = EmptyCh ;
  LocalBuffer [ ConvWidth ] = null ;
  Loop = ConvWidth - 1 ;
  for ( ; ; )
  {
    LastDigit = ( uint8_t ) ( Value % 10 ) ;
    LocalBuffer [ Loop ] = LastDigit + ( uint8_t ) 0x30 ;
    Value = Value / 10 ;
    if ( ( ! Loop ) || ( ! Value ) )
      break ;
    Loop -- ;
  } /* for */ ;
  for ( Loop = 0 ; Loop <= ConvWidth ; Loop ++ )
    * Number ++ = LocalBuffer [ Loop ] ;
} /* WordToStr */


static uint8_t * BegOfStr ( uint8_t * String )
{
  /*----------------------------------------------------------------*/
  while ( * String == Space )
    String ++ ;
  return ( String ) ;
} /* BegOfStr */


static uint8_t StringToNumb ( uint8_t * String )
{
  uint8_t Ch ;
  uint8_t RetValue ;
  /*----------------------------------------------------------------*/
  RetValue = 0 ;
  for ( ; ; )
  {
    Ch = * String ++ ;
    if ( ! Ch )
      break ;
    if ( ( Ch < '0' ) || ( Ch > '9' ) )
    {
      RetValue = 0 ;
      break ;
    } /* if */ ;
    RetValue = 10 * RetValue + Ch - 0x30 ;
  } /* for */ ;
  return ( RetValue ) ;
} /* StringToNumb */


static void EEBlockOperation ( void )
{
  uint16_t CRCValue ;
  /*----------------------------------------------------------------*/
  ReadEEBlock ( ) ;
  if ( EEBlock . EEBlockMetric != StandardSygnature )
  {
    SetDefaultEEBlock ( ) ;
    WriteEEBlock ( ) ;
    return ;
  } /* if */ ;
  CRCValue = ComputeCRC ( sizeof ( EEPROMBlockDataRecT ) ,
                          ( uint8_t * ) ( & EEBlock ) ) ;
  DecodePassword ( ) ;
  if ( CRCValue != 0 )
  {
    SetDefaultEEBlock ( ) ;
    WriteEEBlock ( ) ;
  } /* if */ ;
} /* EEBlockOperation */


static uint8_t MeDestination ( void )
{
  /*----------------------------------------------------------------*/
  return ( CompareDeviceID ( SerialService . DestinationDeviceName , EEBlock . MyDeviceID ) ) ;
} /* MeDestination */


static uint8_t GeneralDestination ( void )
{
  /*----------------------------------------------------------------*/
  return ( CompareFlash ( SerialService . DestinationDeviceName , ( uint16_t ) GeneralPoolDeviceName ) ) ;
} /* GeneralDestination */


static void AddToCyclicReplyDevicesID ( void )
{
  /*----------------------------------------------------------------*/
  AddToCyclicTxt ( SerialService . SourceDeviceName ) ;
  AddToCyclic ( PartSeparator ) ;
  AddToCyclicTxt ( SerialService . DestinationDeviceName ) ;
  AddToCyclic ( PartSeparator ) ;
} /* AddToCyclicReplyDevicesID */


static void AddToCyclicReplyForCommand ( void )
{
  /*----------------------------------------------------------------*/
  AddToCyclic ( 'R' ) ;
  AddToCyclicTxt ( SerialService . CommandText ) ;
  AddToCyclic ( '.' ) ;
} /* AddToCyclicReplyForCommand */


static void AddToCyclicCommandReplyOK ( void )
{
  /*----------------------------------------------------------------*/
  AddToCyclicReplyDevicesID ( ) ;
  AddToCyclicReplyForCommand ( ) ;
  AddToCyclicFlashTxt ( ( uint16_t ) OK ) ;
} /* AddToCyclicCommandReplyOK */


static void AddToCyclicCommandReplyFail ( void )
{
  /*----------------------------------------------------------------*/
  AddToCyclicReplyDevicesID ( ) ;
  AddToCyclicReplyForCommand ( ) ;
  AddToCyclicFlashTxt ( ( uint16_t ) Fail ) ;
} /* AddToCyclicCommandReplyFail */


static void ConnectCommandExecute ( void )
{
  uint8_t Ch ;
  uint8_t Loop ;
  /*----------------------------------------------------------------*/
  for ( Loop = 0 ; Loop <= PasswordSize ; Loop ++ )
  {
    Ch = EEBlock . Password [ Loop ] ;
    if ( SerialService . ParamText [ Loop ] != Ch )
      return ;
    if ( Ch == null )
      break ;
  } /* for */ ;
  ConnectContext = true ;
  AddToCyclicCommandReplyOK ( ) ;
} /* ConnectCommandExecute */


static void DisConnectCommandExecute ( void )
{
  /*----------------------------------------------------------------*/
  ConnectContext = false ;
  AddToCyclicCommandReplyOK ( ) ;
} /* DisConnectCommandExecute */


static void SetDeviceNameCommandExecute ( void )
{
  /*----------------------------------------------------------------*/
  if ( ! ConnectContext )
    return ;
  if ( ! SerialService . ParamText [ 0 ] )
  {
    AddToCyclicCommandReplyFail ( ) ;
    return ;
  } /* if */ ;
  Assign ( EEBlock . MyDeviceID ) ;
  AddToCyclicCommandReplyOK ( ) ;
} /* SetDeviceNameCommandExecute */


static void HelloCommandExecute ( void )
{
  /*----------------------------------------------------------------*/
  AddToCyclicReplyDevicesID ( ) ;
  AddToCyclicReplyForCommand ( ) ;
  AddToCyclicFlashTxt ( ( uint16_t ) DeviceName ) ;
} /* HelloCommandExecute */


static void GetVersionCommandExecute ( void )
{
  /*----------------------------------------------------------------*/
  AddToCyclicReplyDevicesID ( ) ;
  AddToCyclicReplyForCommand ( ) ;
  AddToCyclicFlashTxt ( ( uint16_t ) HardwareSoftwareVersion ) ;
} /* GetVersionCommandExecute */


static void GetIDCommandExecute ( void )
{
  /*----------------------------------------------------------------*/
  AddToCyclicReplyDevicesID ( ) ;
  AddToCyclicReplyForCommand ( ) ;
  AddToCyclicTxt ( EEBlock . MyDeviceID ) ;
  AddToCyclic ( cr ) ;
} /* GetIDCommandExecute */


static void ReinitCommandExecute ( void )
{
  /*----------------------------------------------------------------*/
  if ( ! ConnectContext )
    return ;
  EEBlockOperation ( ) ;
  AddToCyclicCommandReplyOK ( ) ;
} /* ReinitCommandExecute */


static void ResetEECommandExecute ( void )
{
  /*----------------------------------------------------------------*/
  if ( ! ConnectContext )
    return ;
  SetDefaultEEBlock ( ) ;
  AddToCyclicCommandReplyOK ( ) ;
} /* ResetEECommandExecute */


static void SaveEECommandExecute ( void )
{
  uint16_t CRCValue ;
  /*----------------------------------------------------------------*/
  if ( ! ConnectContext )
    return ;
  EEBlock . CRCValue = 0 ;
  CRCValue = ComputeCRC ( sizeof ( EEPROMBlockDataRecT ) -
                          sizeof ( uint16_t ) ,
                          ( uint8_t * ) ( & EEBlock ) ) ;
  EEBlock . CRCValue = CRCValue ;
  WriteEEBlock ( ) ;
  AddToCyclicCommandReplyOK ( ) ;
} /* SaveEECommandExecute */


static void GetChnStatusCommandExecute ( void )
{
  /*----------------------------------------------------------------*/
  AddToCyclicReplyDevicesID ( ) ;
  AddToCyclicReplyForCommand ( ) ;
  AddToCyclicTxt ( EEBlock . MyDeviceID ) ;
  AddToCyclic ( ParamSeparator ) ;
  if ( MotorOnFlag )
  {
    AddToCyclicFlashTxt ( ( uint16_t ) ChannelIsOnState ) ;
  } /* if ... */
  else
  {
    AddToCyclicFlashTxt ( ( uint16_t ) ChannelIsOffState ) ;
  } /* if ... else */ ;
} /* GetChnStatusCommandExecute */


static void AddSuperiorChnCommandExecute ( void )
{
  uint8_t Loop ;
  /*----------------------------------------------------------------*/
  if ( ! ConnectContext )
    return ;
  if ( ! SerialService . ParamText [ 0 ] )
  {
    AddToCyclicCommandReplyFail ( ) ;
    return ;
  } /* if */ ;
  for ( Loop = 0 ; Loop < EventDestinDeviceArraySize ; Loop ++ )
  {
    if ( ! EEBlock . EventDestinDeviceID [ Loop ] [ 0 ] )
    {
      Assign ( EEBlock . EventDestinDeviceID [ Loop ] ) ;
      AddToCyclicCommandReplyOK ( ) ;
      return ;
    } /* if */ ;
  } /* for */ ;
  AddToCyclicCommandReplyFail ( ) ;
} /* AddSuperiorChnCommandExecute */


static void DelSuperiorChnCommandExecute ( void )
{
  uint8_t Loop ;
  /*----------------------------------------------------------------*/
  if ( ! ConnectContext )
    return ;
  if ( ! SerialService . ParamText [ 0 ] )
  {
    AddToCyclicCommandReplyFail ( ) ;
    return ;
  } /* if */ ;
  for ( Loop = 0 ; Loop < EventDestinDeviceArraySize ; Loop ++ )
  {
    if ( CompareDeviceID ( EEBlock . EventDestinDeviceID [ Loop ] , SerialService . ParamText ) )
    {
      ClearBlock ( EEBlock . EventDestinDeviceID [ Loop ] , sizeof ( DeviceIDType ) ) ;
      AddToCyclicCommandReplyOK ( ) ;
      return ;
    } /* if */ ;
  } /* for */ ;
  AddToCyclicCommandReplyFail ( ) ;
} /* DelSuperiorChnCommandExecute */


static void ListSuperiorChnCommandExecute ( void )
{
  uint8_t Loop ;
  uint8_t EmptyTable ;
  /*----------------------------------------------------------------*/
  EmptyTable = true ;
  for ( Loop = 0 ; Loop < EventDestinDeviceArraySize ; Loop ++ )
  {
    if ( EEBlock . EventDestinDeviceID [ Loop ] [ 0 ] )
    {
      EmptyTable = false ;
      break ;
    } /* if */ ;
  } /* for */ ;
  if ( EmptyTable )
  {
    AddToCyclicReplyDevicesID ( ) ;
    AddToCyclicReplyForCommand ( ) ;
    AddToCyclicFlashTxt ( ( uint16_t ) EmptyFlagTxt ) ;
    return ;
  } /* if */ ;
  for ( Loop = 0 ; Loop < EventDestinDeviceArraySize ; Loop ++ )
  {
    if ( EEBlock . EventDestinDeviceID [ Loop ] [ 0 ] )
    {
      AddToCyclicReplyDevicesID ( ) ;
      AddToCyclicReplyForCommand ( ) ;
      AddToCyclicTxt ( EEBlock . EventDestinDeviceID [ Loop ] ) ;
      AddToCyclic ( cr ) ;
    } /* if */ ;
  } /* for */ ;
} /* ListSuperiorChnCommandExecute */


static void ChStateServiceOnCommandExecute ( void )
{
  uint8_t Loop ;
  /*----------------------------------------------------------------*/
  if ( ! SerialService . ParamText [ 0 ] )
  {
    AddToCyclicCommandReplyFail ( ) ;
    return ;
  } /* if */ ;
  for ( Loop = 0 ; Loop < TempEchoDevicesArraySize ; Loop ++ )
  {
    if ( ! TempEchoDevices [ Loop ] [ 0 ] )
    {
      Assign ( TempEchoDevices [ Loop ] ) ;
      AddToCyclicCommandReplyOK ( ) ;
      return ;
    } /* if */ ;
  } /* for */ ;
  AddToCyclicCommandReplyFail ( ) ;
} /* ChStateServiceOnCommandExecute */


static void ChStateServiceOffCommandExecute ( void )
{
  uint8_t Loop ;
  /*----------------------------------------------------------------*/
  if ( ! SerialService . ParamText [ 0 ] )
  {
    AddToCyclicCommandReplyFail ( ) ;
    return ;
  } /* if */ ;
  for ( Loop = 0 ; Loop < TempEchoDevicesArraySize ; Loop ++ )
  {
    if ( CompareDeviceID ( TempEchoDevices [ Loop ] , SerialService . ParamText ) )
    {
      ClearBlock ( TempEchoDevices [ Loop ] , sizeof ( DeviceIDType ) ) ;
      AddToCyclicCommandReplyOK ( ) ;
      return ;
    } /* if */ ;
  } /* for */ ;
  AddToCyclicCommandReplyFail ( ) ;
} /* ChStateServiceOffCommandExecute */


static void ChStateServiceListCommandExecute ( void )
{
  uint8_t Loop ;
  uint8_t EmptyTable ;
  /*----------------------------------------------------------------*/
  EmptyTable = true ;
  for ( Loop = 0 ; Loop < TempEchoDevicesArraySize ; Loop ++ )
  {
    if ( TempEchoDevices [ Loop ] [ 0 ] )
    {
      EmptyTable = false ;
      break ;
    } /* if */ ;
  } /* for */ ;
  if ( EmptyTable )
  {
    AddToCyclicReplyDevicesID ( ) ;
    AddToCyclicReplyForCommand ( ) ;
    AddToCyclicFlashTxt ( ( uint16_t ) EmptyFlagTxt ) ;
    return ;
  } /* if */ ;
  for ( Loop = 0 ; Loop < TempEchoDevicesArraySize ; Loop ++ )
  {
    if ( TempEchoDevices [ Loop ] [ 0 ] )
    {
      AddToCyclicReplyDevicesID ( ) ;
      AddToCyclicReplyForCommand ( ) ;
      AddToCyclicTxt ( TempEchoDevices [ Loop ] ) ;
      AddToCyclic ( cr ) ;
    } /* if */ ;
  } /* for */ ;
} /* ChStateServiceListCommandExecute */


static void SetTempThresholdCommandExecute ( void )
{
  uint16_t NewThresholdValue ;
  /*----------------------------------------------------------------*/
  if ( ! ConnectContext )
    return ;
  if ( ! SerialService . ParamText [ 0 ] )
  {
    AddToCyclicCommandReplyFail ( ) ;
    return ;
  } /* if */ ;
  NewThresholdValue = StringToNumb ( SerialService . ParamText ) ;
  if ( NewThresholdValue )
  {
    EEBlock . TempThreshold = NewThresholdValue ;
    AddToCyclicCommandReplyOK ( ) ;
    return ;
  } /* if */ ;
  AddToCyclicCommandReplyFail ( ) ;
} /* SetTempThresholdCommandExecute */


static void GetTempThresholdCommandExecute ( void )
{
  uint8_t Number [ 8 ] ;
  /*----------------------------------------------------------------*/
  WordToStr ( Number , 5 , Space , EEBlock . TempThreshold ) ;
  AddToCyclicReplyDevicesID ( ) ;
  AddToCyclicReplyForCommand ( ) ;
  AddToCyclicTxt ( BegOfStr ( Number ) ) ;
  AddToCyclic ( cr ) ;
} /* GetTempThresholdCommandExecute */


static void SetTempHysteresisCommandExecute ( void )
{
  uint16_t NewHysteresisValue ;
  /*----------------------------------------------------------------*/
  if ( ! ConnectContext )
    return ;
  if ( ! SerialService . ParamText [ 0 ] )
  {
    AddToCyclicCommandReplyFail ( ) ;
    return ;
  } /* if */ ;
  NewHysteresisValue = StringToNumb ( SerialService . ParamText ) ;
  if ( NewHysteresisValue )
  {
    EEBlock . TempHysteresis = NewHysteresisValue ;
    AddToCyclicCommandReplyOK ( ) ;
    return ;
  } /* if */ ;
  AddToCyclicCommandReplyFail ( ) ;
} /* SetTempHysteresisCommandExecute */


static void GetTempHysteresisCommandExecute ( void )
{
  uint8_t Number [ 8 ] ;
  /*----------------------------------------------------------------*/
  WordToStr ( Number , 5 , Space , EEBlock . TempHysteresis ) ;
  AddToCyclicReplyDevicesID ( ) ;
  AddToCyclicReplyForCommand ( ) ;
  AddToCyclicTxt ( BegOfStr ( Number ) ) ;
  AddToCyclic ( cr ) ;
} /* GetTempHysteresisCommandExecute */


static void AddParamCommandExecute ( void )
{
  /*----------------------------------------------------------------*/
  if ( ! ConnectContext )
    return ;
  if ( ! SerialService . ParamText [ 0 ] )
  {
    AddToCyclicCommandReplyFail ( ) ;
    return ;
  } /* if */ ;
  SerialService . ParamText [ ParamNameSize ] = 0 ;
  ClearBlock ( EEBlock . TempParamName , ParamNameSize + 1 ) ;
  StrCopy ( EEBlock . TempParamName , SerialService . ParamText ) ;
  AddToCyclicCommandReplyOK ( ) ;
} /* AddParamCommandExecute */


static void DelParamCommandExecute ( void )
{
  /*----------------------------------------------------------------*/
  if ( ! ConnectContext )
    return ;
  ClearBlock ( EEBlock . TempParamName , ParamNameSize + 1 ) ;
  AddToCyclicCommandReplyOK ( ) ;
} /* DelParamCommandExecute */


static void ListParamCommandExecute ( void )
{
  /*----------------------------------------------------------------*/
  AddToCyclicReplyDevicesID ( ) ;
  AddToCyclicReplyForCommand ( ) ;
  if ( EEBlock . TempParamName [ 0 ] )
  {
    AddToCyclicTxt ( EEBlock . TempParamName ) ;
    AddToCyclic ( cr ) ;
  } /* if ... */
  else
  {
    AddToCyclicFlashTxt ( ( uint16_t ) EmptyFlagTxt ) ;
  } /* if ... else */ ;
} /* ListParamCommandExecute */


static void NewPasswordCommandExecute ( void )
{
  uint8_t Ch ;
  uint8_t Loop ;
  /*----------------------------------------------------------------*/
  if ( ! ConnectContext )
    return ;
  if ( ! SerialService . ParamText [ 0 ] )
  {
    for ( Loop = 0 ; Loop <= PasswordSize ; Loop ++ )
    {
      Ch = SerialService . ParamText [ Loop ] ;
      EEBlock . Password [ Loop ] = Ch ;
      if ( Ch == null )
        break ;
    } /* for */ ;
    AddToCyclicCommandReplyOK ( ) ;
    return ;
  } /* if */ ;
  AddToCyclicCommandReplyFail ( ) ;
} /* NewPasswordCommandExecute */


static void AddEventText ( uint8_t * DestDeviceID )
{
  /*----------------------------------------------------------------*/
  AddToCyclicTxt ( DestDeviceID ) ;
  AddToCyclic ( PartSeparator ) ;
  AddToCyclicTxt ( EEBlock . MyDeviceID ) ;
  AddToCyclic ( PartSeparator ) ;
  if ( MotorOnFlag )
    AddToCyclicFlashTxt ( ( uint16_t ) ChannelIsOnState ) ;
  else
    AddToCyclicFlashTxt ( ( uint16_t ) ChannelIsOffState ) ;
} /* AddEventText */


static void PermEchoTransmit ( void )
{
  uint8_t Loop ;
  /*----------------------------------------------------------------*/
  for ( Loop = 0 ; Loop < EventDestinDeviceArraySize ; Loop ++ )
  {
    if ( EEBlock . EventDestinDeviceID [ Loop ] [ 0 ] )
    {
      AddEventText ( EEBlock . EventDestinDeviceID [ Loop ] ) ;
    } /* if */ ;
  } /* for */ ;
} /* PermEchoTransmit */


static void TempEchoTransmit ( void )
{
  uint8_t Loop ;
  /*----------------------------------------------------------------*/
  for ( Loop = 0 ; Loop < TempEchoDevicesArraySize ; Loop ++ )
  {
    if ( TempEchoDevices [ Loop ] [ 0 ] )
    {
      AddEventText ( TempEchoDevices [ Loop ] ) ;
    } /* if */ ;
  } /* for */ ;
} /* TempEchoTransmit */


static void MotorOnCommand ( void )
{
  /*----------------------------------------------------------------*/
  if ( StandardControlEnable )
  {
    if ( ! MotorOnFlag )
    {
      MotorOnFlag = true ;
      PowerPort &= ~ ( 1 << PwrSwitchPin ) ;
      LEDPort &= ~ ( 1 << PowerOnOffLed ) ;
      PermEchoTransmit ( ) ;
      TempEchoTransmit ( ) ;
    } /* if */ ;
    if ( MotorOnHandler )
    {
      DetachTimer ( MotorOnHandler ) ;
      MotorOnHandler = NULL ;
    } /* if */ ;
    if ( MotorOffHandler )
    {
      DetachTimer ( MotorOffHandler ) ;
      MotorOffHandler = NULL ;
    } /* if */ ;
  } /* if */ ;
} /* MotorOnCommand */


static void AutoOnMotorAction ( void ) ;

static void AutoOffMotorAction( void )
{
  /*----------------------------------------------------------------*/
  StandardControlEnable = true ;
  MotorOnFlag = false ;
  PowerPort |= ( 1 << PwrSwitchPin ) ;
  LEDPort |= ( 1 << PowerOnOffLed ) ;
  PermEchoTransmit ( ) ;
  TempEchoTransmit ( ) ;
  if ( MotorOnHandler )
  {
    DetachTimer ( MotorOnHandler ) ;
  } /* if */ ;
  MotorOnHandler = AttachTimer ( AutoOnMotorAction , AutoOnMotorTimeInterval ) ;
} /* AutoOffMotorAction */


static void AutoOnMotorAction ( void )
{
  /*----------------------------------------------------------------*/
  StandardControlEnable = false ;
  MotorOnFlag = true ;
  PowerPort &= ~ ( 1 << PwrSwitchPin ) ;
  LEDPort &= ~ ( 1 << PowerOnOffLed ) ;
  PermEchoTransmit ( ) ;
  TempEchoTransmit ( ) ;
  if ( MotorOffHandler )
    DetachTimer ( MotorOffHandler ) ;
  MotorOffHandler = AttachTimer ( AutoOffMotorAction , AutoOffMotorTimeInterval ) ;
} /* AutoOnMotorAction */


static void MotorOffCommand ( void )
{
  /*----------------------------------------------------------------*/
  if ( StandardControlEnable )
  {
    if ( MotorOnFlag )
    {
      MotorOnFlag = false ;
      PowerPort |= ( 1 << PwrSwitchPin ) ;
      LEDPort |= ( 1 << PowerOnOffLed ) ;
      PermEchoTransmit ( ) ;
      TempEchoTransmit ( ) ;
      if ( MotorOnHandler )
        DetachTimer ( MotorOnHandler ) ;
      if ( MotorOffHandler )
        DetachTimer ( MotorOffHandler ) ;
      MotorOnHandler = AttachTimer ( AutoOnMotorAction , AutoOnMotorTimeInterval ) ;
    } /* if */ ;
    MotorOnFlag = false ;
  } /* if */ ;
} /* MotorOffCommand */


static void NewTempMeansCommandExecute ( void )
{
  uint8_t Loop ;
  uint8_t IndexInParamArray ;
  uint8_t Ch ;
  ParamNameType TmpParamName ;
  uint16_t BinaryParamValue ;
  /*----------------------------------------------------------------*/
  if ( ! SerialService . ParamText [ 0 ] )
  {
    AddToCyclicCommandReplyFail ( ) ;
    return ;
  } /* if */ ;
  ClearBlock ( TmpParamName , ParamNameSize + 1 ) ;
  IndexInParamArray = 0 ;
  for ( Loop = 0 ; Loop <= ParamNameSize ; Loop ++ )
  {
    Ch = SerialService . ParamText [ IndexInParamArray ] ;
    IndexInParamArray ++ ;
    if ( ! Ch )
    {
      AddToCyclicCommandReplyFail ( ) ;
      return ;
    } /* if */ ;
    if ( Ch == ParamSeparator )
      break ;
    if ( Loop < ParamNameSize )
      TmpParamName [ Loop ] = Ch ;
  } /* for */ ;
  if ( CompareDeviceID ( EEBlock . TempParamName , TmpParamName ) )
  {
    ClearBlock ( TmpParamName , ParamNameSize + 1 ) ;
    for ( Loop = 0 ; Loop < ParamNameSize ; Loop ++ )
    {
      Ch = SerialService . ParamText [ IndexInParamArray ] ;
      IndexInParamArray ++ ;
      if ( ( ! Ch ) || ( Ch == '.' ) || ( Ch == ',' ) )
        break ;
      TmpParamName [ Loop ] = Ch ;
    } /* for */ ;
    BinaryParamValue = StringToNumb ( TmpParamName ) ;
    if ( BinaryParamValue >= HiTempBorder )
    {
      MotorOnCommand ( ) ;
    } /* if */ ;
    if ( BinaryParamValue <= LoTempBorder )
    {
      MotorOffCommand ( ) ;
    } /* if */ ;
    AddToCyclicCommandReplyOK ( ) ;
    return ;
  } /* if */ ;
  AddToCyclicCommandReplyFail ( ) ;
} /* NewTempMeansCommandExecute */


static void ReplyCommandExecute ( void )
{
  /*----------------------------------------------------------------*/
  if ( CyclicDataPresent ( ) )
  {
    FlushCyclic ( ) ;
  } /* if ... */
  else
  {
    SendSerialTxt ( SerialService . SourceDeviceName ) ;
    SendSerial ( PartSeparator ) ;
    SendSerialTxt ( SerialService . DestinationDeviceName ) ;
    SendSerial ( PartSeparator ) ;
    SendSerialFlashTxt ( ( uint16_t ) NoDataText ) ;
  } /* if ... else */ ;
} /* ReplyCommandExecute */


static void ExecuteAssociatedCommand ( uint8_t *      String ,
                                       uint8_t        AssociationTableSize ,
                                       uint16_t       AssociationTable )
{
  uint8_t Loop ;
  CommandNameAssociationRecT AssociationElement ;
  uint16_t MemAddress ;
  /*----------------------------------------------------------------*/
  MemAddress = AssociationTable ;
  for ( Loop = 0 ; Loop < AssociationTableSize ; Loop ++ )
  {
    AssociationElement . CommandService = ( CommandExecuteServiceProcT ) pgm_read_word_near ( MemAddress ) ;
    MemAddress += 2 ;
    AssociationElement . CommandName = ( uint8_t * ) pgm_read_word_near ( MemAddress ) ;
    MemAddress += 2 ;
    if ( CompareFlash ( String , ( uint16_t ) AssociationElement . CommandName ) )
    {
      AssociationElement . CommandService ( ) ;
      return ;
    } /* if */ ;
  } /* for */ ;
} /* ExecuteAssociatedCommand */


static void ExecuteCommand ( void )
{
  uint8_t FrameToMe ;
  /*----------------------------------------------------------------*/
  FrameToMe = MeDestination ( ) ;
  if ( ! FrameToMe )
  {
    FrameToMe = GeneralDestination ( ) ;
    if ( FrameToMe )
      StrCopy ( SerialService . DestinationDeviceName ,
                EEBlock . MyDeviceID ) ;
  } /* if */ ;
  if ( FrameToMe )
  {
    ExecuteAssociatedCommand ( SerialService . CommandText ,
                               CommandNameAssociationTabeSize ,
                               ( uint16_t ) CommandNameAssociationTabe ) ;
  } /* if */ ;
} /* ExecuteCommand */


static void ResetOperation ( void ) ;

static void DestinationIDPutCh ( uint8_t Ch )
{
  /*----------------------------------------------------------------*/
  if ( SerialService . DestDeviceNameIndex < DeviceIDSize )
  {
    SerialService . DestinationDeviceName [ SerialService . DestDeviceNameIndex ++ ] = Ch ;
  } /* if */ ;
} /* DestinationIDPutCh */


static void SourceIDPutCh ( uint8_t Ch )
{
  /*----------------------------------------------------------------*/
  if ( SerialService . SrcDeviceNameIndex < DeviceIDSize )
  {
    SerialService . SourceDeviceName [ SerialService . SrcDeviceNameIndex ++ ] = Ch ;
  } /* if */ ;
} /* SourceIDPutCh */


static void CommandIDPutCh ( uint8_t Ch )
{
  /*----------------------------------------------------------------*/
  if ( SerialService . CommandTextIndex < CommandBufferSize )
  {
    SerialService . CommandText [ SerialService . CommandTextIndex ++ ] = Ch ;
  } /* if */ ;
} /* CommandIDPutCh */


static void ParamPutCh ( uint8_t Ch )
{
  /*----------------------------------------------------------------*/
  if ( SerialService . ParamTextIndex < ParamBufferSize )
  {
    SerialService . ParamText [ SerialService . ParamTextIndex ++ ] = Ch ;
  } /* if */ ;
} /* ParamPutCh */


static void SelectParam ( void )
{
  uint8_t Ch ;
  /*----------------------------------------------------------------*/
  if ( SerialDataPresent ( ) )
  {
    Ch = GetSerial ( ) ;
    if ( Ch == cr )
    {
      ParamPutCh ( null ) ;
      ExecuteCommand ( ) ;
      ResetOperation ( ) ;
      return ;
    } /* if */;
    if ( Ch < Space )
    {
      ResetOperation ( ) ;
      return ;
    } /* if */ ;
    ParamPutCh ( Ch ) ;
  } /* if */ ;
} /* SelectParam */


static void SelectCommand ( void )
{
  uint8_t Ch ;
  /*----------------------------------------------------------------*/
  if ( SerialDataPresent ( ) )
  {
    Ch = GetSerial ( ) ;
    if ( Ch == cr )
    {
      CommandIDPutCh ( null ) ;
      ExecuteCommand ( ) ;
      ResetOperation ( ) ;
      return ;
    } /* if */;
    if ( Ch < Space )
    {
      ResetOperation ( ) ;
      return ;
    } /* if */ ;
    if ( Ch == ParamSeparator )
    {
      CommandIDPutCh ( null ) ;
      SerialService . SerialService = SelectParam ;
      return ;
    } /* if */;
    CommandIDPutCh ( Ch ) ;
  } /* if */ ;
} /* SelectCommand */


static void SelectSourceService ( void )
{
  uint8_t Ch ;
  /*----------------------------------------------------------------*/
  if ( SerialDataPresent ( ) )
  {
    Ch = GetSerial ( ) ;
    if ( Ch < Space )
    {
      ResetOperation ( ) ;
      return ;
    } /* if */ ;
    if ( Ch == PartSeparator )
    {
      SourceIDPutCh ( null ) ;
      SerialService . SerialService = SelectCommand ;
      return ;
    } /* if */;
    SourceIDPutCh ( Ch ) ;
  } /* if */ ;
} /* SelectSourceService */


static void SelectDestOperation ( void )
{
  uint8_t Ch ;
  /*----------------------------------------------------------------*/
  if ( SerialDataPresent ( ) )
  {
    Ch = GetSerial ( ) ;
    if ( Ch < Space )
    {
      ResetOperation ( ) ;
      return ;
    } /* if */ ;
    if ( Ch == PartSeparator )
    {
      DestinationIDPutCh ( null ) ;
      SerialService . SerialService = SelectSourceService ;
      return ;
    } /* if */;
    DestinationIDPutCh ( Ch ) ;
  } /* if */ ;
} /* SelectDestOperation */


static void ResetOperation ( void )
{
  /*----------------------------------------------------------------*/
  ClearBlock ( SerialService . CommandText , sizeof ( CommandBufferType ) ) ;
  ClearBlock ( SerialService . ParamText , sizeof ( ParamBufferType ) ) ;
  ClearBlock ( SerialService . DestinationDeviceName , sizeof ( DeviceIDType ) ) ;
  ClearBlock ( SerialService . SourceDeviceName , sizeof ( DeviceIDType ) ) ;
  SerialService . CommandTextIndex = 0 ;
  SerialService . ParamTextIndex = 0 ;
  SerialService . DestDeviceNameIndex = 0 ;
  SerialService . SrcDeviceNameIndex = 0 ;
  SerialService . SerialService = SelectDestOperation ;
} /* ResetOperation */


static void SetBorders ( void )
{
  /*----------------------------------------------------------------*/
  HiTempBorder = EEBlock . TempThreshold + EEBlock . TempHysteresis ;
  if ( EEBlock . TempThreshold > EEBlock . TempHysteresis )
  {
    LoTempBorder = EEBlock . TempThreshold - EEBlock . TempHysteresis ;
  } /* if ... */
  else
  {
    LoTempBorder = 0 ;
  } /* if ... else */ ;
} /* SetBorders */


static void SoftwareInit ( void )
{
  uint8_t Loop ;
  /*----------------------------------------------------------------*/
  ConnectContext = false ;
  ServiceDelayTime = 0 ;
  TimerCounter = 0 ;
  CyclicBuffer . CycFr = 0 ;
  CyclicBuffer . CycTo = 0 ;
  SerialService . RXDRec . CycFr = 0 ;
  SerialService . RXDRec . CycTo = 0 ;
  SerialService . TXDRec . CycFr = 0 ;
  SerialService . TXDRec . CycTo = 0 ;
  SerialService . TXDRec . State = IdleState ;
  SerialService . SerialLEDDelay = 0 ;
  for ( Loop = 0 ; Loop < TempEchoDevicesArraySize ; Loop ++ )
  {
    ClearBlock ( TempEchoDevices [ Loop ] , sizeof ( DeviceIDType ) ) ;
  } /* for */ ;
  ResetOperation ( ) ;
  TimeEvent = false ;
  MotorOnFlag = false ;
  MotorOnHandler = NULL ;
  MotorOffHandler = NULL ;
  SoftTimeEventInit ( ) ;
  StandardControlEnable = true ;
} /* SoftwareInit */


int main ( void )
{
  /*----------------------------------------------------------------*/
  HardwareInit ( Speed38400bps ) ;
  EnvirInit ( ) ;
  SoftwareInit ( ) ;
  USARTReceiveMode ( ) ;
  EEBlockOperation ( ) ;
  wdt_enable ( WDTO_1S ) ;
  sei ( ) ;
  SetBorders ( ) ;
  for ( ; ; )
  {
    SysTimerPool ( ) ;
    wdt_reset ( ) ;
    if ( TimeEvent )
    {
      TimeEvent = false ;
      LEDPort ^= 1 << StatusLed ;
    } /* if */ ;
    if ( SerialService . SerialService )
      SerialService . SerialService ( ) ;
  } /* for */ ;
  return ( 0 ) ;
} /* main */


